/*
 * WPA Supplicant - driver interaction with Devicescape stack
 * Copyright (c) 2003-2005, Sven Henkel <shenkel@gmail.com>
 *
 * This file includes code from the hostap-driver for WPA supplicant
 * Copyright (c) 2003-2005, Jouni Malinen <jkmaline@cc.hut.fi>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Alternatively, this software may be distributed under the terms of BSD
 * license.
 *
 * See README and COPYING for more details.
 */

#include "includes.h"
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <errno.h>

#include "wireless_copy.h"
#include "common.h"
#include "driver.h"
#include "driver_wext.h"
#include "eloop.h"
#include "driver_dscape.h"




struct wpa_driver_dscape_data {
        void *wext; /* private data for driver_wext */
        void *ctx;
        char ifname[IFNAMSIZ + 1];
        int sock;
        int current_mode; /* infra/adhoc */
};


static int dscape_ioctl(struct wpa_driver_dscape_data *drv,
			 struct prism2_hostapd_param *param,
			 int len, int show_err)
{
	struct iwreq iwr;

	memset(&iwr, 0, sizeof(iwr));
	strncpy(iwr.ifr_name, drv->ifname, IFNAMSIZ);
	iwr.u.data.pointer = (caddr_t) param;
	iwr.u.data.length = len;

	if (ioctl(drv->sock, PRISM2_IOCTL_HOSTAPD, &iwr) < 0) {
		int ret = errno;
		if (show_err) 
			perror("ioctl[PRISM2_IOCTL_HOSTAPD]");
		return ret;
	}

	return 0;
}

static int prism2param(struct wpa_driver_dscape_data *drv, int param,
		       int value)
{
	struct iwreq iwr;
	int *i, ret = 0;

	memset(&iwr, 0, sizeof(iwr));
	strncpy(iwr.ifr_name, drv->ifname, IFNAMSIZ);
	i = (int *) iwr.u.name;
	*i++ = param;
	*i++ = value;

	if (ioctl(drv->sock, PRISM2_IOCTL_PRISM2_PARAM, &iwr) < 0) {
		fprintf(stderr, "ioctl[PRISM2_IOCTL_PRISM2_PARAM] param = %d  value = %d\n", param, value);
		ret = -1;
	}
	return ret;
}

static int wpa_driver_dscape_set_wpa_ie(struct wpa_driver_dscape_data *drv, const u8 *wpa_ie, size_t wpa_ie_len)
{
	struct prism2_hostapd_param *param;
	int res;
	size_t blen = PRISM2_DSCAPE_GENERIC_ELEMENT_HDR_LEN + wpa_ie_len;

	if (blen < sizeof(*param))
		blen = sizeof(*param);

	param = (struct prism2_hostapd_param *) malloc(blen);
	if (param == NULL)
		return -1;

	memset(param, 0, blen);
	param->cmd = PRISM2_HOSTAPD_SET_GENERIC_INFO_ELEM;
	param->u.set_generic_info_elem.len = wpa_ie_len;
	memcpy(param->u.set_generic_info_elem.data, wpa_ie, wpa_ie_len);
	res = dscape_ioctl(drv, param, blen, 1);

	free(param);

	return res;
}

static int wpa_driver_dscape_set_wpa(void *priv, int enabled)
{
	struct wpa_driver_dscape_data *drv = priv;
	int ret = 0;

	wpa_printf(MSG_DEBUG, "%s: enabled=%d", __FUNCTION__, enabled);

	if (!enabled && wpa_driver_dscape_set_wpa_ie(drv, NULL, 0) < 0)
		ret = -1;

	//if (prism2param(drv, PRISM2_PARAM_HOST_ROAMING, enabled ? 2 : 0) < 0)
		//ret = -1;
	//if (prism2param(drv, PRISM2_PARAM_WPA, enabled) < 0)
		//ret = -1;

	return ret;
}



static int wpa_driver_dscape_set_key(void *priv, wpa_alg alg, const u8 *addr, int key_idx,
					int set_tx, const u8 *seq, size_t seq_len,
					const u8 *key, size_t key_len)
{
	struct wpa_driver_dscape_data *drv = priv;
	struct prism2_hostapd_param *param;

	u8 *buf;
	size_t blen;
	int ret = 0;
	char *alg_name;
	//int i;

	switch (alg) {
	case WPA_ALG_NONE:
		alg_name = "none";
		break;
	case WPA_ALG_WEP:
		alg_name = "WEP";
		break;
	case WPA_ALG_TKIP:
		alg_name = "TKIP";
		break;
	case WPA_ALG_CCMP:
		alg_name = "CCMP";
		break;
	default:
		return -1;
	}

	wpa_printf(MSG_DEBUG, "%s: alg=%s key_idx=%d set_tx=%d seq_len=%lu "
		   "key_len=%lu", __FUNCTION__, alg_name, key_idx, set_tx,
		   (unsigned long) seq_len, (unsigned long) key_len);

// 	printf("key = ");
// 	for (i = 0; i < key_len; i++) {
// 		printf("%.2x ", key[i]);
// 	}
// 	printf("\n");

	if (seq_len > 8)
		return -2;

	blen = sizeof(*param) + key_len;
	buf = malloc(blen);
	if (buf == NULL)
		return -1;
	memset(buf, 0, blen);

	param = (struct prism2_hostapd_param *) buf;
	param->cmd = PRISM2_SET_ENCRYPTION;
	/* TODO: In theory, STA in client mode can use five keys; four default
	 * keys for receiving (with keyidx 0..3) and one individual key for
	 * both transmitting and receiving (keyidx 0) _unicast_ packets. Now,
	 * keyidx 0 is reserved for this unicast use and default keys can only
	 * use keyidx 1..3 (i.e., default key with keyidx 0 is not supported).
	 * This should be fine for more or less all cases, but for completeness
	 * sake, the driver could be enhanced to support the missing key. */
#if 0 
	if (addr == NULL)
		memset(param->sta_addr, 0xff, ETH_ALEN);
	else
		memcpy(param->sta_addr, addr, ETH_ALEN);
#else
	memset(param->sta_addr, 0xff, ETH_ALEN);
#endif
	strncpy((char *) param->u.crypt.alg, alg_name,
		HOSTAP_CRYPT_ALG_NAME_LEN);
	param->u.crypt.flags = set_tx ? HOSTAP_CRYPT_FLAG_SET_TX_KEY : 0;
	param->u.crypt.idx = key_idx;
	//memcpy(param->u.crypt.seq, seq, seq_len);
	param->u.crypt.key_len = key_len;
	memcpy(param->u.crypt.key, key, key_len);

	if (dscape_ioctl(drv, param, blen, 1)) {
		wpa_printf(MSG_WARNING, "Failed to set encryption.");
		//show_set_key_error(param);
		ret = -1;
	}
	free(buf);

	return ret;
}

static int wpa_driver_dscape_set_countermeasures(void *priv, int enabled)
{
	//FIXME: we should tell the driver to drop all tkip packets iff enabled
	fprintf(stderr, "FIXME wpa_driver_dscape_set_countermeasures entered\n");
	return 0;
}

static int wpa_driver_dscape_set_drop_unencrypted(void *priv, int enabled)
{
	struct wpa_driver_dscape_data *drv = priv;
	int ret = prism2param(drv, PRISM2_PARAM_DROP_UNENCRYPTED, enabled);
	if (ret < 0)
		return ret;

	// allow eapol packets
	ret = prism2param(drv, PRISM2_PARAM_EAPOL, enabled);
	if (ret < 0)
		return ret;

	return 0;
}

static int wpa_driver_dscape_mlme(struct wpa_driver_dscape_data *drv, const u8 *addr, int cmd, int reason_code)
{
        struct prism2_hostapd_param param;
        int ret;

        /* There does not seem to be a better way of deauthenticating or
         * disassociating with Prism2/2.5/3 than sending the management frame
         * and then resetting the Port0 to make sure both the AP and the STA
         * end up in disconnected state. */
        memset(&param, 0, sizeof(param));
        param.cmd = PRISM2_HOSTAPD_MLME;
        memcpy(param.sta_addr, addr, ETH_ALEN);
        param.u.mlme.cmd = cmd;
        param.u.mlme.reason_code = reason_code;
        ret = dscape_ioctl(drv, &param, sizeof(param), 1);

//        if (ret == 0) {
//                usleep(100000);
//                ret = wpa_driver_hostap_reset(drv, 2);
//        }
        return ret;
}

static int wpa_driver_hostap_associate(void *priv, struct wpa_driver_associate_params *params)
{
	struct wpa_driver_dscape_data *drv = priv;
	int ret = 0;
	int ieee8021x = 0;
	int key_mgmt = IEEE80211_KEY_MGMT_NONE;

	wpa_printf(MSG_DEBUG, "%s", __FUNCTION__);

// 	if (params->mode != drv->current_mode) {
// 		/* At the moment, Host AP driver requires host_roaming=2 for
// 		 * infrastructure mode and host_roaming=0 for adhoc. */
// 		if (prism2param(drv, PRISM2_PARAM_HOST_ROAMING,
// 				params->mode == IEEE80211_MODE_IBSS ? 0 : 2) <
// 		    0) {
// 			wpa_printf(MSG_DEBUG, "%s: failed to set host_roaming",
// 				   __func__);
// 		}
// 		drv->current_mode = params->mode;
// 	}

	if (prism2param(drv, PRISM2_PARAM_PRIVACY_INVOKED,
			params->key_mgmt_suite != KEY_MGMT_NONE) < 0)
		ret = -1;
	if (wpa_driver_dscape_set_wpa_ie(drv, params->wpa_ie,
					 params->wpa_ie_len) < 0)
		ret = -1;

	switch(params->key_mgmt_suite) {
		case KEY_MGMT_802_1X:
			ieee8021x = 1;
			key_mgmt = IEEE80211_KEY_MGMT_IEEE8021X;
			break;
		case KEY_MGMT_PSK:
			key_mgmt = IEEE80211_KEY_MGMT_WPA_PSK;
			break;
		case KEY_MGMT_NONE:
			key_mgmt = IEEE80211_KEY_MGMT_NONE;
			break;
		case KEY_MGMT_802_1X_NO_WPA:
		case KEY_MGMT_WPA_NONE:
			key_mgmt = IEEE80211_KEY_MGMT_NONE; //FIXME Do those modes have a corresponding mode in dscape?
			break;
	}

	if (prism2param(drv, PRISM2_PARAM_IEEE_802_1X, ieee8021x) < 0) {
		wpa_printf(MSG_DEBUG, "hostap: Failed to configure "
			   "ieee_802_1x param");
		ret = -1;
	}

	if (prism2param(drv, PRISM2_PARAM_KEY_MGMT, key_mgmt) < 0) {
		wpa_printf(MSG_DEBUG, "hostap: Failed to configure "
			   "key-management param");
		ret = -1;
	}

	if (wpa_driver_wext_set_mode(drv->wext, params->mode) < 0)
		ret = -1;
	if (params->freq &&
	    wpa_driver_wext_set_freq(drv->wext, params->freq) < 0)
		ret = -1;
	if (wpa_driver_wext_set_ssid(drv->wext, params->ssid, params->ssid_len)
	    < 0)
		ret = -1;
	if (wpa_driver_wext_set_bssid(drv->wext, params->bssid) < 0)
		ret = -1;

	return ret;
}


static int wpa_driver_dscape_deauthenticate(void *priv, const u8 *addr, int reason_code)
{
	struct wpa_driver_dscape_data *drv = priv;
        wpa_printf(MSG_DEBUG, "%s", __FUNCTION__);
        return wpa_driver_dscape_mlme(drv, addr, MLME_STA_DEAUTH, reason_code);
}

static int wpa_driver_dscape_disassociate(void *priv, const u8 *addr, int reason_code)
{
	struct wpa_driver_dscape_data *drv = priv;
        wpa_printf(MSG_DEBUG, "%s", __FUNCTION__);
        return wpa_driver_dscape_mlme(drv, addr, MLME_STA_DISASSOC, reason_code);
}

static int wpa_driver_dscape_scan(void *priv, const u8 *ssid, size_t ssid_len)
{
	struct wpa_driver_dscape_data *drv = priv;
	struct prism2_hostapd_param param;
	int ret;

	fprintf(stderr, "wpa_driver_dscape_scan entered\n");

	if (ssid == NULL) {
		/* Use standard Linux Wireless Extensions ioctl if possible
		 * because some drivers using hostap code in wpa_supplicant
		 * might not support Host AP specific scan request (with SSID
		 * info). */
		return wpa_driver_wext_scan(drv->wext, ssid, ssid_len);
	}

	if (ssid_len > 32)
		ssid_len = 32;

	memset(&param, 0, sizeof(param));
	param.cmd = PRISM2_HOSTAPD_SCAN_REQ;
	param.u.scan_req.ssid_len = ssid_len;
	memcpy(param.u.scan_req.ssid, ssid, ssid_len);
	ret = dscape_ioctl(drv, &param, sizeof(param), 1);

	/* Not all drivers generate "scan completed" wireless event, so try to
	 * read results after a timeout. */
	eloop_register_timeout(3, 0, wpa_driver_wext_scan_timeout, drv->wext,
			       drv->ctx);

	return ret;
}

static int wpa_driver_dscape_set_auth_alg(void *priv, int auth_alg)
{
	struct wpa_driver_dscape_data *drv = priv;
	int algs = 0;

	if (auth_alg & AUTH_ALG_OPEN_SYSTEM)
		algs |= 1;
	if (auth_alg & AUTH_ALG_SHARED_KEY)
		algs |= 2;
	if (auth_alg & AUTH_ALG_LEAP)
		algs |= 4;
	if (algs == 0)
		algs = 1; /* at least one algorithm should be set */

	return prism2param(drv, PRISM2_PARAM_AP_AUTH_ALGS, algs);
	return 0;
}

static int wpa_driver_dscape_get_bssid(void *priv, u8 *bssid)
{
	struct wpa_driver_dscape_data *drv = priv;
	return wpa_driver_wext_get_bssid(drv->wext, bssid);
}

static int wpa_driver_dscape_get_ssid(void *priv, u8 *ssid)
{
	struct wpa_driver_dscape_data *drv = priv;
	return wpa_driver_wext_get_ssid(drv->wext, ssid);
}

static int wpa_driver_dscape_get_scan_results(void *priv, struct wpa_scan_result *results, size_t max_size)
{
	struct wpa_driver_dscape_data *drv = priv;
	return wpa_driver_wext_get_scan_results(drv->wext);
}

static void * wpa_driver_dscape_init(void *ctx, const char *ifname)
{
	struct wpa_driver_dscape_data *drv;

	drv = malloc(sizeof(*drv));
	if (drv == NULL)
		return NULL;
	memset(drv, 0, sizeof(*drv));

	drv->wext = wpa_driver_wext_init(ctx, ifname);
	if (drv->wext == NULL) {
                free(drv);
                return NULL;
        }

        drv->ctx = ctx;
        strncpy(drv->ifname, ifname, sizeof(drv->ifname));
        drv->sock = socket(PF_INET, SOCK_DGRAM, 0);
        if (drv->sock < 0) {
                perror("socket");
                wpa_driver_wext_deinit(drv->wext);
                free(drv);
                return NULL;
        }

	return drv;
}

static void wpa_driver_dscape_deinit(void *priv)
{
	struct wpa_driver_dscape_data *drv = priv;
        wpa_driver_wext_deinit(drv->wext);
        close(drv->sock);
        free(drv);
}

const struct wpa_driver_ops wpa_driver_dscape_ops = {
        .name = "dscape",
        .desc = "Devicescape driver",
        .get_bssid = wpa_driver_dscape_get_bssid,
        .get_ssid = wpa_driver_dscape_get_ssid,
        .set_wpa = wpa_driver_dscape_set_wpa,
        .set_key = wpa_driver_dscape_set_key,
        .set_countermeasures = wpa_driver_dscape_set_countermeasures,
        .set_drop_unencrypted = wpa_driver_dscape_set_drop_unencrypted,
        .scan = wpa_driver_dscape_scan,
        .get_scan_results = wpa_driver_dscape_get_scan_results,
        .deauthenticate = wpa_driver_dscape_deauthenticate,
        .disassociate = wpa_driver_dscape_disassociate,
        .associate = wpa_driver_hostap_associate,
        .set_auth_alg = wpa_driver_dscape_set_auth_alg,
        .init = wpa_driver_dscape_init,
        .deinit = wpa_driver_dscape_deinit,
};
